/* PV264 2020 – HW01: Graph Iterators
 *
 * Your task is to implement a set of iterators over a directed graph. The order
 * in which the graph's vertices are iterated corresponds to the well-known
 * algorithms BFS and DFS. For DFS, implement both pre-order and post-order
 * variants.
 *
 * Moreover, all three iterators come in two flavours:
 *   «sourced» search, which starts in a specified vertex and iterates through
 *             all the vertices reachable from that vertex
 *   «global» search, which traverses the entire graph by starting the search
 *            in every theretofore undiscovered vertex.
 *
 * Of course, there is a constant and a mutable variant of each of these six
 * iterators. Implementing these 12 iterators without code duplication is
 * a large part of this assignment.
 *
 * The iterators must compute the graph traversal on-the-fly – pre-computing the
 * right order is not a viable solution. Take extra care to implement the
 * searches efficiently (in terms of both time and space complexity) and
 * correctly (so that your DFS is indeed a DFS).
 *
 * Edges from each vertex shall be explored in the order in which they were
 * defined using a call to ‹add_edge›. Similarly, the global iterators shall
 * start the search in all the vertices in the order of their respective
 * ‹add_node› calls. Note that during the global search no vertex is visited
 * twice (i.e., the set of visited vertices is shared across the searches).
 * Call of ‹add_edge› or ‹add_node› invalidates all associated iterators.
 *
 * The internal graph representation is up to you; just provide a ‹node_handle›
 * type. Its values must be cheap to copy and must remain valid after the graph
 * has been moved. Please note that assignment does not enforce that the vertex
 * value must be unique, nor that there is at most one edge between vertices.
 *
 * When designing the iterators, refer to the “legacy iterator requirements” on
 * <https://en.cppreference.com/w/cpp/iterator>. The iterators must satisfy
 * InputIterator (the ones returned by the non-const member functions also
 * OutputIterator) with the following restrictions (taken from ForwardIterator):
 *   - incrementing an iterator doesn't affect its copies nor what gets read
 *     from (or modified through) them.
 *   - assignment through a mutable iterator doesn't invalidate it or its
 *     copies.
 *   - ‹a == b› implies ‹++a == ++b›. (Not really interesting.)
 *
 * In contrast to ForwardIterator, the ‹==› relation doesn't have to distinguish
 * between all the pointed-to vertices. It only has to detect the end of the
 * traversal and be an equality relation. However, feel free to implement
 * a full-blown ForwardIterator; it should emerge naturally.
 *
 * Don't forget to make sure that ‹std::iterator_traits› of your iterators
 * contain the correct types.
 *
 * When implementing the iterators, do not duplicate code (i.e. there should be
 * only one implementation of BFS and DFS (try not to duplicate code between the
 * pre-order and post-order variants, either) and only one implementation of the
 * sourced and the global strategy). The constant and mutable variants should
 * share basically all the code. Try to make everything about the iterator
 * flavours compile-time: «virtual functions are forbidden altogether» and you
 * should avoid using various (run-time) flags as well.
 *
 * Templates are your friends. A few facts that some of the approaches use:
 *
 *   - A class may inherit from its own template parameter. Calling the parent's
 *     member functions then requires an explicit ‹this->›.
 *   - When you call a member function of a templated object, the existence of
 *     the member function is checked «during instantiation» – no need to know
 *     the exact interface beforehand.
 *   - You may call ‹static› member functions of template parameters as ‹T::foo()›.
 *   - Not only a type, but also an integral constant (such as a number or an
 *     ‹enum› member) may be a template parameter.
 *   - You can template a ‹using› declaration to partially specialise a template
 *     under a new name.
 *   - You can use ‹using Base::Base;› inside a class derived from ‹Base› to
 *     re-export all constructors of the parent class as the derived one's own.
 *
 * For comparison, both reference solutions have under 250 lines.
 */

 /** You can assume that V is movable **/
template< typename V >
struct graph {
    using value_type = V;
    using node_handle = /** TODO **/;

    node_handle add_node( V );
    void add_edge( node_handle, node_handle );

    /** Explicitly read-only **/
    auto bfs_cbegin( node_handle ) const;
    auto bfs_cend( node_handle ) const;

    auto bfs_cbegin() const;
    auto bfs_cend() const;

    auto dfs_preorder_cbegin( node_handle ) const;
    auto dfs_preorder_cend( node_handle ) const;

    auto dfs_preorder_cbegin() const;
    auto dfs_preorder_cend() const;

    auto dfs_postorder_cbegin( node_handle ) const;
    auto dfs_postorder_cend( node_handle ) const;

    auto dfs_postorder_cbegin() const;
    auto dfs_postorder_cend() const;

    /** Implicitly read-only **/
    auto bfs_begin( node_handle ) const;
    auto bfs_end( node_handle ) const;

    auto bfs_begin() const;
    auto bfs_end() const;

    auto dfs_preorder_begin( node_handle ) const;
    auto dfs_preorder_end( node_handle ) const;

    auto dfs_preorder_begin() const;
    auto dfs_preorder_end() const;

    auto dfs_postorder_begin( node_handle ) const;
    auto dfs_postorder_end( node_handle ) const;

    auto dfs_postorder_begin() const;
    auto dfs_postorder_end() const;

    /** Read-write **/
    auto bfs_begin( node_handle );
    auto bfs_end( node_handle );

    auto bfs_begin();
    auto bfs_end();

    auto dfs_preorder_begin( node_handle );
    auto dfs_preorder_end( node_handle );

    auto dfs_preorder_begin();
    auto dfs_preorder_end();

    auto dfs_postorder_begin( node_handle );
    auto dfs_postorder_end( node_handle );

    auto dfs_postorder_begin();
    auto dfs_postorder_end();
};
