#include "graph_iterators.hpp"

#include <cassert>
#include <vector>
#include <memory>
#include <iterator>

/** Here we provide you some basic tests, so that you get some basic
 ** feeling about how the iterators can be used/will be tested.
 ** You are encouraged to write your own,
 ** provided tests here are but a small subset of sanity tests! **/

template< typename It, typename V >
void iter_properties()
{
    static_assert( !std::is_polymorphic_v< It >,
                   "Virtual member functions are forbidden in this assignment."  );
    static_assert( std::is_copy_constructible_v< It > );
    static_assert( std::is_copy_assignable_v< It > );
    // ...
}

template< typename It, typename V >
void it_traits()
{
    using Traits = std::iterator_traits< It >;

    static_assert( std::is_same_v< typename Traits::value_type, std::remove_cv_t< V > > );
    static_assert( std::is_same_v< typename Traits::pointer, V * > );
    static_assert( std::is_base_of_v< std::input_iterator_tag, typename Traits::iterator_category > );
    // ...
}

template< typename It, typename V >
void it_static_properties()
{
    iter_properties< It, V >();
    it_traits< It, V >();
}


template< typename Arg, typename ... Args >
void check_iter_static()
{
    graph< Arg > g0;
    it_static_properties< decltype( g0.bfs_begin() ), Arg >();
    it_static_properties< decltype( g0.bfs_cend() ), const Arg >();
    it_static_properties< decltype( g0.dfs_postorder_begin() ), Arg >();
    it_static_properties< decltype( g0.dfs_preorder_begin() ), Arg >();
    it_static_properties< decltype( g0.dfs_preorder_cbegin() ), const Arg >();
    it_static_properties< decltype( g0.dfs_preorder_cbegin() ), const Arg > ();
    if constexpr ( sizeof ... ( Args ) != 0 )
        check_iter_static< Args ... >();
}

template< typename It >
void basic_iter_methods( It begin, It end )
{
    assert( *begin );
    assert( begin == begin );
    assert( begin != end );
    assert( end != begin );

    auto old = begin;
    assert( *begin == *old );
    assert( begin == old );
    assert( *old != *(++begin) );
    assert( ( begin = old ) == old );
    assert( *(++begin) != *old );

    // ...
}

template< typename G >
void basic_iter_methods( G &g )
{
    basic_iter_methods( g.bfs_begin(), g.bfs_end() );
    basic_iter_methods( g.bfs_cbegin(), g.bfs_cend() );

    basic_iter_methods( g.dfs_postorder_begin(), g.dfs_postorder_end() );
    // ...
}

/** Copy all encountered values into vector. To avoid some more advanced C++ techniques this
 ** function requires that `It::value_type` is copyable (for example `std::unique_ptr` is not).
 **/
template< typename It >
auto collect( It begin, It end )
{
    std::vector< typename It::value_type > out;
    while ( begin != end )
        out.push_back( *begin++ );
    return out;
}

int main()
{
    check_iter_static< int, std::string, std::unique_ptr< int >, std::optional< std::string > >();

    graph< int > g0;
    auto v4 = g0.add_node( 4 );
    auto v3 = g0.add_node( 3 );
    auto v2 = g0.add_node( 2 );
    g0.add_edge( v4, v3 );
    g0.add_edge( v3, v4 );
    g0.add_edge( v3, v2 );

    basic_iter_methods( g0 );

    using vi_t = std::vector< int >;
    assert( collect( g0.bfs_begin(), g0.bfs_end() ) == collect( g0.bfs_cbegin(), g0.bfs_cend() ) );
    assert( collect( g0.bfs_begin(), g0.bfs_end() ) == vi_t( { 4, 3, 2 } ) );
    assert( collect( g0.bfs_begin( v3 ), g0.bfs_end( v3 ) ) == vi_t( { 3, 4, 2 } ) );
    // ...

    graph< std::unique_ptr< int > > g1;
    auto v42 = g1.add_node( std::make_unique< int >( 42 ) );
    auto v12 = g1.add_node( std::make_unique< int >( 12 ) );
    basic_iter_methods( g1 );

    assert( ++g1.dfs_preorder_cbegin( v42 ) == g1.dfs_preorder_cend( v42 ) );
    assert( ++g1.dfs_preorder_cbegin( v12 ) == g1.dfs_preorder_cend( v12 ) );

    // ...
}
